@@ -12,6 +12,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 from abc import ABC, abstractproperty
 from collections import OrderedDict
 
@@ -29,9 +30,12 @@ from .calendar_helpers import (
     compute_all_minutes,
     next_divider_idx,
     previous_divider_idx,
+    Session,
+    parse_session,
 )
 from .utils.memoize import lazyval
 from .utils.pandas_utils import days_at_time
+from .pandas_extensions.offsets import MultipleWeekmaskCustomBusinessDay
 
 start_default = pd.Timestamp("1990-01-01", tz=UTC)
 end_base = pd.Timestamp("today", tz=UTC)
@@ -115,6 +119,9 @@ class ExchangeCalendar(ABC):
             self.close_offset,
         )
 
+        # Apply special offsets first
+        self._calculate_and_overwrite_special_offsets(_all_days, start, end)
+
         # `Series`s mapping sessions with nonstandard opens/closes to
         # the open/close time.
         _special_opens = self._calculate_special_opens(start, end)
@@ -134,14 +141,22 @@ class ExchangeCalendar(ABC):
             _special_closes,
         )
 
+        if self._break_starts is None:
+            break_starts = None
+        else:
+            break_starts = self._break_starts.tz_localize(None)
+        if self._break_ends is None:
+            break_ends = None
+        else:
+            break_ends = self._break_ends.tz_localize(None)
         self.schedule = DataFrame(
             index=_all_days,
             data=OrderedDict(
                 [
-                    ("market_open", self._opens),
-                    ("break_start", self._break_starts),
-                    ("break_end", self._break_ends),
-                    ("market_close", self._closes),
+                    ("market_open", self._opens.tz_localize(None)),
+                    ("break_start", break_starts),
+                    ("break_end", break_ends),
+                    ("market_close", self._closes.tz_localize(None)),
                 ]
             ),
             dtype="datetime64[ns]",
@@ -152,28 +167,20 @@ class ExchangeCalendar(ABC):
         # with the same inputs.
         self._minute_to_session_label_cache = (None, None)
 
-        self.market_opens_nanos = self.schedule.market_open.values.astype(
+        self.market_opens_nanos = self.schedule.market_open.values.astype(np.int64)
+
+        self.market_break_starts_nanos = self.schedule.break_start.values.astype(
             np.int64
         )
 
-        self.market_break_starts_nanos = (
-            self.schedule.break_start.values.astype(np.int64)
-        )
+        self.market_break_ends_nanos = self.schedule.break_end.values.astype(np.int64)
 
-        self.market_break_ends_nanos = self.schedule.break_end.values.astype(
-            np.int64
-        )
-
-        self.market_closes_nanos = self.schedule.market_close.values.astype(
-            np.int64
-        )
+        self.market_closes_nanos = self.schedule.market_close.values.astype(np.int64)
 
         _check_breaks_match(
             self.market_break_starts_nanos, self.market_break_ends_nanos
         )
 
-        self._trading_minutes_nanos = self.all_minutes.values.astype(np.int64)
-
         self.first_trading_session = _all_days[0]
         self.last_trading_session = _all_days[-1]
 
@@ -187,14 +194,22 @@ class ExchangeCalendar(ABC):
 
     @lazyval
     def day(self):
-        return CustomBusinessDay(
-            holidays=self.adhoc_holidays,
-            calendar=self.regular_holidays,
-            weekmask=self.weekmask,
-        )
+        if self.special_weekmasks:
+            return MultipleWeekmaskCustomBusinessDay(
+                holidays=self.adhoc_holidays,
+                calendar=self.regular_holidays,
+                weekmask=self.weekmask,
+                weekmasks=self.special_weekmasks,
+            )
+        else:
+            return CustomBusinessDay(
+                holidays=self.adhoc_holidays,
+                calendar=self.regular_holidays,
+                weekmask=self.weekmask,
+            )
 
     @abstractproperty
-    def name(self):
+    def name(self) -> str:
         raise NotImplementedError()
 
     @abstractproperty
@@ -258,16 +273,16 @@ class ExchangeCalendar(ABC):
 
     @lazyval
     def _minutes_per_session(self):
-        close_to_open_diff = (
-            self.schedule.market_close - self.schedule.market_open
+        close_to_open_diff = self.schedule.market_close - self.schedule.market_open
+        break_diff = (self.schedule.break_end - self.schedule.break_start).fillna(
+            pd.Timedelta(seconds=0)
         )
-        break_diff = (
-            self.schedule.break_end - self.schedule.break_start
-        ).fillna(pd.Timedelta(seconds=0))
         diff = (close_to_open_diff - break_diff).astype("timedelta64[m]")
         return diff + 1
 
-    def minutes_count_for_sessions_in_range(self, start_session, end_session):
+    def minutes_count_for_sessions_in_range(
+        self, start_session: pd.Timestamp, end_session: pd.Timestamp
+    ) -> int:
         """
         Parameters
         ----------
@@ -299,7 +314,7 @@ class ExchangeCalendar(ABC):
         """
         Returns
         -------
-        list : A list of timestamps representing unplanned closes.
+        list: A list of timestamps representing unplanned closes.
         """
         return []
 
@@ -345,32 +360,83 @@ class ExchangeCalendar(ABC):
         """
         return []
 
+    @property
+    def special_weekmasks(self):
+        """
+        Returns
+        -------
+        list: List of (date, date, str) tuples that represent special
+         weekmasks that applies between dates.
+        """
+        return []
+
+    @property
+    def special_offsets(self):
+        """
+        Returns
+        -------
+        list: List of (timedelta, timedelta, timedelta, timedelta, AbstractHolidayCalendar) tuples
+         that represent special open, break_start, break_end, close offsets
+         and corresponding HolidayCalendars.
+        """
+        return []
+
+    @property
+    def special_offsets_adhoc(self):
+        """
+        Returns
+        -------
+        list: List of (timedelta, timedelta, timedelta, timedelta, DatetimeIndex) tuples
+         that represent special open, break_start, break_end, close offsets
+         and corresponding DatetimeIndexes.
+        """
+        return []
+
     # -----
 
     @property
-    def opens(self):
+    def opens(self) -> pd.Series:
         return self.schedule.market_open
 
     @property
-    def closes(self):
+    def closes(self) -> pd.Series:
         return self.schedule.market_close
 
     @property
-    def late_opens(self):
+    def late_opens(self) -> pd.DatetimeIndex:
         return self._late_opens
 
     @property
-    def early_closes(self):
+    def early_closes(self) -> pd.DatetimeIndex:
         return self._early_closes
 
     @property
-    def break_starts(self):
+    def break_starts(self) -> pd.Series:
         return self.schedule.break_start
 
     @property
-    def break_ends(self):
+    def break_ends(self) -> pd.Series:
         return self.schedule.break_end
 
+    @lazyval
+    def all_minutes(self) -> pd.DatetimeIndex:
+        """
+        Returns a DatetimeIndex representing all the minutes in this calendar.
+        """
+        return DatetimeIndex(
+            compute_all_minutes(
+                self.market_opens_nanos,
+                self.market_break_starts_nanos,
+                self.market_break_ends_nanos,
+                self.market_closes_nanos,
+            ),
+            tz=UTC,
+        )
+
+    @lazyval
+    def all_minutes_nanos(self) -> np.ndarray:
+        return self.all_minutes.values.astype(np.int64)
+
     def is_session(self, dt):
         """
         Given a dt, returns whether it's a valid session label.
@@ -387,6 +453,32 @@ class ExchangeCalendar(ABC):
         """
         return dt in self.schedule.index
 
+    def has_breaks(
+        self,
+        start: Session | None = None,
+        end: Session | None = None,
+    ) -> bool:
+        """Query if at least one session of a calendar has a break.
+
+        Parameters
+        ----------
+        start : optional
+            Limit query to sessions from `start`.
+
+        end : optional
+            Limit query to sessions through `end`.
+
+        Returns
+        -------
+        bool
+            True if any calendar session has a break, false otherwise.
+        """
+        if start is not None:
+            start = parse_session(start, "start", self, strict=False)
+        if end is not None:
+            end = parse_session(end, "end", self, strict=False)
+        return self.break_starts[start:end].notna().any()
+
     def is_open_on_minute(self, dt, ignore_breaks=False):
         """
         Given a dt, return whether this exchange is open at the given dt.
@@ -415,9 +507,7 @@ class ExchangeCalendar(ABC):
             if ignore_breaks:
                 return True
 
-            break_start_on_open_dt = self.market_break_starts_nanos[
-                open_idx - 1
-            ]
+            break_start_on_open_dt = self.market_break_starts_nanos[open_idx - 1]
             break_end_on_open_dt = self.market_break_ends_nanos[open_idx - 1]
             # NaT comparisions will result in False
             if break_start_on_open_dt < dt < break_end_on_open_dt:
@@ -522,7 +612,7 @@ class ExchangeCalendar(ABC):
         pd.Timestamp
             The next exchange minute.
         """
-        idx = next_divider_idx(self._trading_minutes_nanos, dt.value)
+        idx = next_divider_idx(self.all_minutes_nanos, dt.value)
         return self.all_minutes[idx]
 
     def previous_minute(self, dt):
@@ -542,10 +632,10 @@ class ExchangeCalendar(ABC):
             The previous exchange minute.
         """
 
-        idx = previous_divider_idx(self._trading_minutes_nanos, dt.value)
+        idx = previous_divider_idx(self.all_minutes_nanos, dt.value)
         return self.all_minutes[idx]
 
-    def next_session_label(self, session_label):
+    def next_session_label(self, session_label: Session) -> pd.Timestamp:
         """
         Given a session label, returns the label of the next session.
 
@@ -564,6 +654,7 @@ class ExchangeCalendar(ABC):
         Raises ValueError if the given session is the last session in this
         calendar.
         """
+        session_label = parse_session(session_label, "session_label", self)
         idx = self.schedule.index.get_loc(session_label)
         try:
             return self.schedule.index[idx + 1]
@@ -576,7 +667,7 @@ class ExchangeCalendar(ABC):
             else:
                 raise
 
-    def previous_session_label(self, session_label):
+    def previous_session_label(self, session_label: Session) -> pd.Timestamp:
         """
         Given a session label, returns the label of the previous session.
 
@@ -595,16 +686,16 @@ class ExchangeCalendar(ABC):
         Raises ValueError if the given session is the first session in this
         calendar.
         """
+        session_label = parse_session(session_label, "session_label", self)
         idx = self.schedule.index.get_loc(session_label)
         if idx == 0:
             raise ValueError(
                 "There is no previous session as this is the"
                 " beginning of the exchange calendar."
             )
-
         return self.schedule.index[idx - 1]
 
-    def minutes_for_session(self, session_label):
+    def minutes_for_session(self, session_label: Session) -> int:
         """
         Given a session label, return the minutes for that session.
 
@@ -618,12 +709,15 @@ class ExchangeCalendar(ABC):
         pd.DateTimeIndex
             All the minutes for the given session.
         """
+        session_label = parse_session(session_label, "session_label", self)
         return self.minutes_in_range(
             start_minute=self.schedule.at[session_label, "market_open"],
             end_minute=self.schedule.at[session_label, "market_close"],
         )
 
-    def execution_minutes_for_session(self, session_label):
+    def execution_minutes_for_session(
+        self, session_label: Session
+    ) -> pd.DatetimeIndex:
         """
         Given a session label, return the execution minutes for that session.
 
@@ -637,6 +731,7 @@ class ExchangeCalendar(ABC):
         pd.DateTimeIndex
             All the execution minutes for the given session.
         """
+        session_label = parse_session(session_label, "session_label", self)
         return self.minutes_in_range(
             start_minute=self.execution_time_from_open(
                 self.schedule.at[session_label, "market_open"],
@@ -650,25 +745,21 @@ class ExchangeCalendar(ABC):
         minutes = self.execution_minutes_for_session
         return pd.DatetimeIndex(
             np.concatenate(
-                [
-                    minutes(session)
-                    for session in self.sessions_in_range(start, stop)
-                ]
+                [minutes(session) for session in self.sessions_in_range(start, stop)]
             ),
             tz=UTC,
         )
 
-    def minutes_window(self, start_dt, count):
+    def minutes_window(self, start_dt: pd.Timestamp, count: int):
         start_dt_nanos = start_dt.value
-        all_minutes_nanos = self._trading_minutes_nanos
-        start_idx = all_minutes_nanos.searchsorted(start_dt_nanos)
+        start_idx = self.all_minutes_nanos.searchsorted(start_dt_nanos)
 
         # searchsorted finds the index of the minute **on or after** start_dt.
         # If the latter, push back to the prior minute.
-        if all_minutes_nanos[start_idx] != start_dt_nanos:
+        if self.all_minutes_nanos[start_idx] != start_dt_nanos:
             start_idx -= 1
 
-        if start_idx < 0 or start_idx >= len(all_minutes_nanos):
+        if start_idx < 0 or start_idx >= len(self.all_minutes_nanos):
             raise KeyError("Can't start minute window at {}".format(start_dt))
 
         end_idx = start_idx + count
@@ -697,12 +788,12 @@ class ExchangeCalendar(ABC):
             The desired sessions.
         """
         return self.all_sessions[
-            self.all_sessions.slice_indexer(
-                start_session_label, end_session_label
-            )
+            self.all_sessions.slice_indexer(start_session_label, end_session_label)
         ]
 
-    def sessions_window(self, session_label, count):
+    def sessions_window(
+        self, session_label: Session, count: int
+    ) -> pd.DatetimeIndex:
         """
         Given a session label and a window size, returns a list of sessions
         of size `count` + 1, that either starts with the given session
@@ -722,14 +813,17 @@ class ExchangeCalendar(ABC):
         pd.DatetimeIndex
             The desired sessions.
         """
+        session_label = parse_session(session_label, "session_label", self)
         start_idx = self.schedule.index.get_loc(session_label)
         end_idx = start_idx + count
         end_idx = max(0, end_idx)
-        return self.all_sessions[
-            min(start_idx, end_idx) : max(start_idx, end_idx) + 1
-        ]
+        return self.all_sessions[min(start_idx, end_idx) : max(start_idx, end_idx) + 1]
 
-    def session_distance(self, start_session_label, end_session_label):
+    def session_distance(
+        self,
+        start_session_label: pd.Timestamp,
+        end_session_label: pd.Timestamp,
+    ):
         """
         Given a start and end session label, returns the distance between them.
         For example, for three consecutive sessions Mon., Tues., and Wed,
@@ -786,21 +880,20 @@ class ExchangeCalendar(ABC):
         pd.DatetimeIndex
             The minutes in the desired range.
         """
-        start_idx = searchsorted(
-            self._trading_minutes_nanos, start_minute.value
-        )
+        start_idx = searchsorted(self.all_minutes_nanos, start_minute.value)
+        end_idx = searchsorted(self.all_minutes_nanos, end_minute.value)
 
-        end_idx = searchsorted(self._trading_minutes_nanos, end_minute.value)
-
-        if end_minute.value == self._trading_minutes_nanos[end_idx]:
+        if end_minute.value == self.all_minutes_nanos[end_idx]:
             # if the end minute is a market minute, increase by 1
             end_idx += 1
 
         return self.all_minutes[start_idx:end_idx]
 
     def minutes_for_sessions_in_range(
-        self, start_session_label, end_session_label
-    ):
+        self,
+        start_session_label: Session,
+        end_session_label: Session,
+    ) -> pd.DatetimeIndex:
         """
         Returns all the minutes for all the sessions from the given start
         session label to the given end session label, inclusive.
@@ -819,12 +912,20 @@ class ExchangeCalendar(ABC):
             The minutes in the desired range.
 
         """
+        start_session_label = parse_session(
+            start_session_label, "start_session_label", self
+        )
+        end_session_label = parse_session(
+            end_session_label, "end_session_label", self
+        )
         first_minute, _ = self.open_and_close_for_session(start_session_label)
         _, last_minute = self.open_and_close_for_session(end_session_label)
 
         return self.minutes_in_range(first_minute, last_minute)
 
-    def open_and_close_for_session(self, session_label):
+    def open_and_close_for_session(
+        self, session_label: Session
+    ) -> tuple(pd.Timestamp, pd.Timestamp):
         """
         Returns a tuple of timestamps of the open and close of the session
         represented by the given label.
@@ -839,12 +940,15 @@ class ExchangeCalendar(ABC):
         (Timestamp, Timestamp)
             The open and close for the given session.
         """
+        session_label = parse_session(session_label, "session_label", self)
         return (
             self.session_open(session_label),
             self.session_close(session_label),
         )
 
-    def break_start_and_end_for_session(self, session_label):
+    def break_start_and_end_for_session(
+        self, session_label: Session
+    ) -> tuple(pd.Timestamp, pd.Timestamp):
         """
         Returns a tuple of timestamps of the break start and end of the session
         represented by the given label.
@@ -859,15 +963,18 @@ class ExchangeCalendar(ABC):
         (Timestamp, Timestamp)
             The break start and end for the given session.
         """
+        session_label = parse_session(session_label, "session_label", self)
         return (
             self.session_break_start(session_label),
             self.session_break_end(session_label),
         )
 
-    def session_open(self, session_label):
+    def session_open(self, session_label: Session) -> pd.Timestamp:
+        session_label = parse_session(session_label, "session_label", self)
         return self.schedule.at[session_label, "market_open"].tz_localize(UTC)
 
-    def session_break_start(self, session_label):
+    def session_break_start(self, session_label: Session) -> pd.Timestamp:
+        session_label = parse_session(session_label, "session_label", self)
         break_start = self.schedule.at[session_label, "break_start"]
         if not pd.isnull(break_start):
             # older versions of pandas need this guard
@@ -875,7 +982,8 @@ class ExchangeCalendar(ABC):
 
         return break_start
 
-    def session_break_end(self, session_label):
+    def session_break_end(self, session_label: Session) -> pd.Timestamp:
+        session_label = parse_session(session_label, "session_label", self)
         break_end = self.schedule.at[session_label, "break_end"]
         if not pd.isnull(break_end):
             # older versions of pandas need this guard
@@ -883,7 +991,8 @@ class ExchangeCalendar(ABC):
 
         return break_end
 
-    def session_close(self, session_label):
+    def session_close(self, session_label: Session) -> pd.Timestamp:
+        session_label = parse_session(session_label, "session_label", self)
         return self.schedule.at[session_label, "market_close"].tz_localize(UTC)
 
     def session_opens_in_range(self, start_session_label, end_session_label):
@@ -899,48 +1008,49 @@ class ExchangeCalendar(ABC):
         ].dt.tz_localize(UTC)
 
     @property
-    def all_sessions(self):
+    def all_sessions(self) -> pd.DatetimeIndex:
         return self.schedule.index
 
     @property
-    def first_session(self):
+    def first_session(self) -> pd.Timestamp:
         return self.all_sessions[0]
 
     @property
-    def last_session(self):
+    def last_session(self) -> pd.Timestamp:
         return self.all_sessions[-1]
 
     @property
-    def first_session_open(self):
+    def first_session_open(self) -> pd.Timestamp:
         """Open time of calendar's first session."""
         return self.opens[0]
 
     @property
-    def last_session_close(self):
+    def last_session_close(self) -> pd.Timestamp:
         """Close time of calendar's last session."""
         return self.closes[-1]
 
+    def session_has_break(self, session: Session) -> bool:
+        """Query if a given session has a break.
+
+        Parameters
+        ----------
+        session :
+            Session to query.
+
+        Returns
+        -------
+        bool
+            True if `session` has a break, false otherwise.
+        """
+        session = parse_session(session, "session", self)
+        return pd.notna(self.session_break_start(session))
+
     def execution_time_from_open(self, open_dates):
         return open_dates
 
     def execution_time_from_close(self, close_dates):
         return close_dates
 
-    @lazyval
-    def all_minutes(self):
-        """
-        Returns a DatetimeIndex representing all the minutes in this calendar.
-        """
-        return DatetimeIndex(
-            compute_all_minutes(
-                self.market_opens_nanos,
-                self.market_break_starts_nanos,
-                self.market_break_ends_nanos,
-                self.market_closes_nanos,
-            ),
-            tz=UTC,
-        )
-
     def minute_to_session_label(self, dt, direction="next"):
         """
         Given a minute, get the label of its containing session.
@@ -1017,13 +1127,14 @@ class ExchangeCalendar(ABC):
                 )
         else:
             # invalid direction
-            raise ValueError(
-                "Invalid direction parameter: " "{0}".format(direction)
-            )
+            raise ValueError("Invalid direction parameter: " "{0}".format(direction))
 
         return current_or_next_session
 
-    def minute_index_to_session_labels(self, index):
+    def minute_index_to_session_labels(
+        self,
+        index: pd.DatetimeIndex | pd.Series,
+    ) -> pd.DatetimeIndex:
         """
         Given a sorted DatetimeIndex of market minutes, return a
         DatetimeIndex of the corresponding session labels.
@@ -1044,12 +1155,8 @@ class ExchangeCalendar(ABC):
             )
         # Find the indices of the previous open and the next close for each
         # minute.
-        prev_opens = (
-            self._opens.values.searchsorted(index.values, side="right") - 1
-        )
-        next_closes = self._closes.values.searchsorted(
-            index.values, side="left"
-        )
+        prev_opens = self._opens.values.searchsorted(index.values, side="right") - 1
+        next_closes = self._closes.values.searchsorted(index.values, side="left")
 
         # If they don't match, the minute is outside the trading day. Barf.
         mismatches = prev_opens != next_closes
@@ -1152,6 +1259,137 @@ class ExchangeCalendar(ABC):
             end,
         )
 
+    def _overwrite_special_offsets(
+        self,
+        session_labels,
+        opens_or_closes,
+        calendars,
+        ad_hoc_dates,
+        start_date,
+        end_date,
+        strict=False,
+    ):
+        # Short circuit when nothing to apply.
+        if opens_or_closes is None or not len(opens_or_closes):
+            return
+
+        len_m, len_oc = len(session_labels), len(opens_or_closes)
+        if len_m != len_oc:
+            raise ValueError(
+                "Found misaligned dates while building calendar.\n"
+                "Expected session_labels to be the same length as "
+                "open_or_closes but,\n"
+                "len(session_labels)=%d, len(open_or_closes)=%d" % (len_m, len_oc)
+            )
+
+        regular = []
+        for offset, calendar in calendars:
+            days = calendar.holidays(start_date, end_date)
+            series = pd.Series(
+                index=pd.DatetimeIndex(days, tz=UTC),
+                data=offset,
+            )
+            regular.append(series)
+
+        ad_hoc = []
+        for offset, datetimes in ad_hoc_dates:
+            series = pd.Series(
+                index=pd.to_datetime(datetimes, utc=True),
+                data=offset,
+            )
+            ad_hoc.append(series)
+
+        merged = regular + ad_hoc
+        if not merged:
+            return pd.Series([], dtype="timedelta64[ns]")
+
+        result = pd.concat(merged).sort_index()
+        offsets = result.loc[(result.index >= start_date) & (result.index <= end_date)]
+
+        # Find the array indices corresponding to each special date.
+        indexer = session_labels.get_indexer(offsets.index)
+
+        # -1 indicates that no corresponding entry was found.  If any -1s are
+        # present, then we have special dates that doesn't correspond to any
+        # trading day.
+        if -1 in indexer and strict:
+            bad_dates = list(offsets.index[indexer == -1])
+            raise ValueError("Special dates %s are not trading days." % bad_dates)
+
+        special_opens_or_closes = opens_or_closes[indexer] + offsets
+
+        # Short circuit when nothing to apply.
+        if not len(special_opens_or_closes):
+            return
+
+        # NOTE: This is a slightly dirty hack.  We're in-place overwriting the
+        # internal data of an Index, which is conceptually immutable.  Since we're
+        # maintaining sorting, this should be ok, but this is a good place to
+        # sanity check if things start going haywire with calendar computations.
+        opens_or_closes.values[indexer] = special_opens_or_closes.values
+
+    def _calculate_and_overwrite_special_offsets(self, session_labels, start, end):
+        _special_offsets = self.special_offsets
+        _special_offsets_adhoc = self.special_offsets_adhoc
+
+        _special_open_offsets = [
+            (t[0], t[-1]) for t in _special_offsets if t[0] is not None
+        ]
+        _special_open_offsets_adhoc = [
+            (t[0], t[-1]) for t in _special_offsets_adhoc if t[0] is not None
+        ]
+        _special_break_start_offsets = [
+            (t[1], t[-1]) for t in _special_offsets if t[1] is not None
+        ]
+        _special_break_start_offsets_adhoc = [
+            (t[1], t[-1]) for t in _special_offsets_adhoc if t[1] is not None
+        ]
+        _special_break_end_offsets = [
+            (t[2], t[-1]) for t in _special_offsets if t[2] is not None
+        ]
+        _special_break_end_offsets_adhoc = [
+            (t[2], t[-1]) for t in _special_offsets_adhoc if t[2] is not None
+        ]
+        _special_close_offsets = [
+            (t[3], t[-1]) for t in _special_offsets if t[3] is not None
+        ]
+        _special_close_offsets_adhoc = [
+            (t[3], t[-1]) for t in _special_offsets_adhoc if t[3] is not None
+        ]
+
+        self._overwrite_special_offsets(
+            session_labels,
+            self._opens,
+            _special_open_offsets,
+            _special_open_offsets_adhoc,
+            start,
+            end,
+        )
+        self._overwrite_special_offsets(
+            session_labels,
+            self._break_starts,
+            _special_break_start_offsets,
+            _special_break_start_offsets_adhoc,
+            start,
+            end,
+        )
+        self._overwrite_special_offsets(
+            session_labels,
+            self._break_ends,
+            _special_break_end_offsets,
+            _special_break_end_offsets_adhoc,
+            start,
+            end,
+        )
+        self._overwrite_special_offsets(
+            session_labels,
+            self._closes,
+            _special_close_offsets,
+            _special_close_offsets_adhoc,
+            start,
+            end,
+        )
+
 
 def _check_breaks_match(market_break_starts_nanos, market_break_ends_nanos):
     """Checks that market_break_starts_nanos and market_break_ends_nanos
@@ -1169,12 +1407,13 @@ def _check_breaks_match(market_break_starts_nanos, market_break_ends_nanos):
             """
             Mismatched market breaks
             Break starts:
-            %s
+            {0}
             Break ends:
-            %s
-            """,
-            market_break_starts_nanos[~nats_match],
-            market_break_ends_nanos[~nats_match],
+            {1}
+            """.format(
+                market_break_starts_nanos[~nats_match],
+                market_break_ends_nanos[~nats_match],
+            )
         )
 
 
@@ -1191,9 +1430,7 @@ def scheduled_special_times(calendar, start, end, time, tz):
     )
 
 
-def _overwrite_special_dates(
-    session_labels, opens_or_closes, special_opens_or_closes
-):
+def _overwrite_special_dates(session_labels, opens_or_closes, special_opens_or_closes):
     """
     Overwrite dates in open_or_closes with corresponding dates in
     special_opens_or_closes, using session_labels for alignment.
@@ -1248,8 +1485,7 @@ def _remove_breaks_for_special_dates(
         raise ValueError(
             "Found misaligned dates while building calendar.\n"
             "Expected session_labels to be the same length as break_starts,\n"
-            "but len(session_labels)=%d, len(break_start_or_end)=%d"
-            % (len_m, len_oc)
+            "but len(session_labels)=%d, len(break_start_or_end)=%d" % (len_m, len_oc)
         )
 
     # Find the array indices corresponding to each special date.
